#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
Handy functions for use in various places.
"""
from ._logger import logger

logger.debug(f'Loading module {__name__}.')

from typing import (
    Any,
    Union,
    Mapping,
    KeysView,
    Generator,
    ItemsView,
    ValuesView,
)
from collections.abc import Iterable
import re

from treelib import Tree
import networkx as nx
import stringcase

__all__ = ['LIST_LIKE', 'SIMPLE_TYPES', 'ensure_iterable', 'to_list']

SIMPLE_TYPES = (
    bytes,
    str,
    int,
    float,
    bool,
    type(None),
)

LIST_LIKE = (
    list,
    set,
    tuple,
    Generator,
    ItemsView,
    KeysView,
    Mapping,
    ValuesView,
)


def to_list(value: Any) -> list:
    """
    Ensures that ``value`` is a list.
    """

    if isinstance(value, LIST_LIKE):

        value = list(value)

    else:

        value = [value]

    return value


def ensure_iterable(value: Any) -> Iterable:
    """
    Returns iterables, except strings, wraps simple types into tuple.
    """

    return value if isinstance(value, LIST_LIKE) else (value, )


def create_tree_visualisation(inheritance_tree: Union[dict, nx.Graph]) -> str:
    """
    Creates a visualisation of the inheritance tree using treelib.
    """

    if isinstance(inheritance_tree, nx.Graph):

        inheritance_tree = nx.to_dict_of_lists(inheritance_tree)
        # unlist values
        inheritance_tree = {k: v[0] for k, v in inheritance_tree.items() if v}

    # find root node
    classes = set(inheritance_tree.keys())
    parents = set(inheritance_tree.values())
    root = list(parents - classes)

    if len(root) > 1:

        if 'entity' in root:

            root = 'entity'  # default: good standard? TODO

        else:

            raise ValueError(
                'Inheritance tree cannot have more than one root node. '
                f'Found {len(root)}: {root}.'
            )

    else:

        root = root[0]

    if not root:
        # find key whose value is None
        root = list(inheritance_tree.keys())[list(inheritance_tree.values()
                                                 ).index(None)]

    tree = Tree()

    tree.create_node(root, root)

    while classes:

        for child in classes:

            parent = inheritance_tree[child]

            if parent in tree.nodes.keys() or parent == root:

                tree.create_node(child, child, parent=parent)

        for node in tree.nodes.keys():

            if node in classes:

                classes.remove(node)

    return tree


# string conversion, adapted from Biolink Model Toolkit
lowercase_pattern = re.compile(r'[a-zA-Z]*[a-z][a-zA-Z]*')
underscore_pattern = re.compile(r'(?<!^)(?=[A-Z][a-z])')


def from_pascal(s: str, sep: str = ' ') -> str:
    underscored = underscore_pattern.sub(sep, s)
    lowercased = lowercase_pattern.sub(
        lambda match: match.group(0).lower(),
        underscored,
    )
    return lowercased


def pascalcase_to_sentencecase(s: str) -> str:
    """
    Convert PascalCase to sentence case.

    Args:
        s: Input string in PascalCase

    Returns:
        string in sentence case form
    """
    return from_pascal(s, sep=' ')


def snakecase_to_sentencecase(s: str) -> str:
    """
    Convert snake_case to sentence case.

    Args:
        s: Input string in snake_case

    Returns:
        string in sentence case form
    """
    return stringcase.sentencecase(s).lower()


def sentencecase_to_snakecase(s: str) -> str:
    """
    Convert sentence case to snake_case.

    Args:
        s: Input string in sentence case

    Returns:
        string in snake_case form
    """
    return stringcase.snakecase(s).lower()


def sentencecase_to_pascalcase(s: str) -> str:
    """
    Convert sentence case to PascalCase.

    Args:
        s: Input string in sentence case

    Returns:
        string in PascalCase form
    """
    return re.sub(r'(?:^| )([a-zA-Z])', lambda match: match.group(1).upper(), s)



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
BioCypher: a unifying framework for biomedical knowledge graphs.
"""

__all__ = [
    '__version__',
    '__author__',
    'module_data',
    'config',
    'logfile',
    'log',
    'Driver',
    'BioCypher',
]

from ._core import BioCypher
from ._config import config, module_data
from ._logger import log, logger, logfile
from ._metadata import __author__, __version__


class Driver(BioCypher):

    # initialise parent class but log a warning
    def __init__(self, *args, **kwargs):
        logger.warning(
            'The class `Driver` is deprecated and will be removed in a future '
            'release. Please use `BioCypher` instead.'
        )
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['biocypher', 'biocypher._config']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.0',
 'appdirs',
 'more_itertools',
 'neo4j-utils==0.0.7',
 'networkx>=3.0,<4.0',
 'rdflib>=6.2.0,<7.0.0',
 'stringcase>=1.2.0,<2.0.0',
 'treelib>=1.6.1,<2.0.0']

setup_kwargs = {
    'name': 'biocypher',
    'version': '0.5.4',
    'description': 'A unifying framework for biomedical research knowledge graphs',
    'long_description': '# BioCypher\n\n## 📖 Documentation\nTutorial and developer docs at https://biocypher.org. For a quickstart into your\nown pipeline, you can refer to our [project\ntemplate](https://github.com/saezlab/biocypher-project-template), and for an \noverview of existing and planned adapters for resources and outputs, as well\nas other features, visit our [GitHub Project \nBoard](https://github.com/orgs/saezlab/projects/5/views/6).\n\n## ❓ Description\nThis is the development repository for BioCypher, our proposal for a unifying\nframework to create knowledge graph databases for systems biomedicine. For an\noverview, usage notes, and a tutorial, read the docs\n[here](https://biocypher.org).\n\n<img\n    style="display: block;\n           margin-left: auto;\n           margin-right: auto;\n           width: 70%;"\n    src="graphical_abstract.png"\n    alt="Graphical Abstract">\n</img>\n\n## ⚙️ Usage\nBioCypher is currently in prototype stage. Installation instructions can be\nfound [here](https://saezlab.github.io/BioCypher/installation.html).\n\nExemplary usage of BioCypher to build a graph database is shown in various\nadapters we have created. For instance, interested users can refer to the\n[migration of the Clinical Knowledge Graph](https://github.com/saezlab/CKG-BioCypher), or\nthe [adapters created for the CROssBAR project](https://github.com/HUBioDataLab/CROssBAR-BioCypher-Migration/blob/main/scripts/create_crossbar.py).\nAs the project evolves, more of these examples will be directly included in the documentation.\n\n## 🤝 Getting involved\nWe are very happy about contributions from the community, large and small!\nIf you would like to contribute to BioCypher development, please refer to\nour [contribution guidelines](CONTRIBUTING.md). :)\n\nIf you want to ask informal questions, talk about dev things, or just chat, please join our community at https://biocypher.zulipchat.com!\n\n> **Imposter syndrome disclaimer:** We want your help. No, really. There may be a little voice inside your head that is telling you that you\'re not ready, that you aren\'t skilled enough to contribute. We assure you that the little voice in your head is wrong. Most importantly, there are many valuable ways to contribute besides writing code.\n>\n> This disclaimer was adapted from the [Pooch](https://github.com/fatiando/pooch) project.\n\n## ✍️ Citation\nThe BioCypher paper is available as a preprint at https://arxiv.org/abs/2212.13543.\n\n## Acknowledgements\nThis project has received funding from the European Union’s Horizon 2020 research and innovation programme under grant agreement No 965193 for DECIDER and No 116030 for TransQST.\n\n[![Powered by the Bioregistry](https://img.shields.io/static/v1?label=Powered%20by&message=Bioregistry&color=BA274A&style=flat&logo=image/png;base64,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)](https://github.com/biopragmatics/bioregistry)\n',
    'author': 'Sebastian Lobentanzer',
    'author_email': 'sebastian.lobentanzer@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/saezlab/biocypher',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
BioCypher 'ontology' module. Contains classes and functions to handle parsing
and representation of single ontologies as well as their hybridisation and
other advanced operations.
"""
import os

from ._logger import logger

logger.debug(f'Loading module {__name__}.')

from typing import Optional
from datetime import datetime

import rdflib
import networkx as nx

from . import _misc
from ._mapping import OntologyMapping


class OntologyAdapter:
    """
    Class that represents an ontology to be used in the Biocypher framework. Can
    read from a variety of formats, including OWL, OBO, and RDF/XML. The
    ontology is represented by a networkx.DiGraph object; an RDFlib graph is
    also kept. By default, the DiGraph reverses the label and identifier of the
    nodes, such that the node name in the graph is the human-readable label. The
    edges are oriented from child to parent. Going from the Biolink example,
    labels are formatted in lower sentence case. In some cases, this means that
    we replace underscores with spaces.
    """
    def __init__(
        self,
        ontology_file,
        root_label,
        head_join_node: Optional[str] = None,
        reverse_labels=True,
        remove_prefixes=True,
    ):
        """
        Initialize the OntologyAdapter class.

        Args:
            ontology_file (str): Path to the ontology file. Can be local or
                remote.

            root_label (str): The label of the root node in the ontology.

            head_join_node (str): Optional variable to store the label of the
                node in the head ontology that should be used to join to the
                root node of the tail ontology. Defaults to None.

            reverse_labels (bool): If True, the node names in the graph will be
                the human-readable labels. If False, the node names will be the
                identifiers. Defaults to True.

            remove_prefixes (bool): If True, the prefixes of the identifiers will
                be removed. Defaults to True.
        """

        logger.info(f'Instantiating OntologyAdapter class for {ontology_file}.')

        self._ontology_file = ontology_file
        self._root_label = root_label
        self._head_join_node = head_join_node
        self._reverse_labels = reverse_labels
        self._remove_prefixes = remove_prefixes

        # Load the ontology into an rdflib Graph according to the file extension
        self._rdf_graph = self._load_rdf_graph(ontology_file)

        self._nx_graph = self._rdf_to_nx(
            self._rdf_graph, root_label, reverse_labels
        )

    def _rdf_to_nx(self, g, root_label, switch_id_and_label=True):

        # Loop through all labels in the ontology
        for s, _, o in g.triples((None, rdflib.RDFS.label, None)):
            # If the label is the root label, set the root node to the subject of the label
            if o.eq(root_label):
                root = s
                break
        else:
            raise ValueError(
                f'Could not find root node with label {root_label}'
            )

        # Create a directed graph to represent the ontology as a tree
        G = nx.DiGraph()

        # Define a recursive function to add subclasses to the graph
        def add_subclasses(node):

            # Only add nodes that have a label
            if (node, rdflib.RDFS.label, None) not in g:
                return

            nx_id, nx_label = _get_nx_id_and_label(node)

            if nx_id not in G:
                G.add_node(nx_id)
                G.nodes[nx_id]['label'] = nx_label

            # Recursively add all subclasses of the node to the graph
            for s, _, o in g.triples((None, rdflib.RDFS.subClassOf, node)):

                # Only add nodes that have a label
                if (s, rdflib.RDFS.label, None) not in g:
                    continue

                s_id, s_label = _get_nx_id_and_label(s)
                G.add_node(s_id)
                G.nodes[s_id]['label'] = s_label

                G.add_edge(s_id, nx_id)
                add_subclasses(s)
                add_parents(s)

        def add_parents(node):

            # Only add nodes that have a label
            if (node, rdflib.RDFS.label, None) not in g:
                return

            nx_id, nx_label = _get_nx_id_and_label(node)

            # Recursively add all parents of the node to the graph
            for s, _, o in g.triples((node, rdflib.RDFS.subClassOf, None)):

                # Only add nodes that have a label
                if (o, rdflib.RDFS.label, None) not in g:
                    continue

                o_id, o_label = _get_nx_id_and_label(o)

                # Skip nodes already in the graph
                if o_id in G:
                    continue

                G.add_node(o_id)
                G.nodes[o_id]['label'] = o_label

                G.add_edge(nx_id, o_id)
                add_parents(o)

        def _get_nx_id_and_label(node):
            node_id_str = self._remove_prefix(str(node))
            node_label_str = str(g.value(node,
                                         rdflib.RDFS.label)).replace('_', ' ')

            nx_id = node_label_str if switch_id_and_label else node_id_str
            nx_label = node_id_str if switch_id_and_label else node_label_str
            return nx_id, nx_label

        # Add all subclasses of the root node to the graph
        add_subclasses(root)

        return G

    def _remove_prefix(self, uri: str) -> str:
        """
        Remove the prefix of a URI. URIs can contain either "#" or "/" as a
        separator between the prefix and the local name. The prefix is
        everything before the last separator.
        """
        if self._remove_prefixes:
            return uri.rsplit('#', 1)[-1].rsplit('/', 1)[-1]
        else:
            return uri

    def _load_rdf_graph(self, ontology_file):
        """
        Load the ontology into an RDFlib graph. The ontology file can be in
        OWL, OBO, or RDF/XML format.
        """
        g = rdflib.Graph()
        g.parse(ontology_file, format=self._get_format(ontology_file))
        return g

    def _get_format(self, ontology_file):
        """
        Get the format of the ontology file.
        """
        if ontology_file.endswith('.owl'):
            return 'application/rdf+xml'
        elif ontology_file.endswith('.obo'):
            raise NotImplementedError('OBO format not yet supported')
        elif ontology_file.endswith('.rdf'):
            return 'application/rdf+xml'
        elif ontology_file.endswith('.ttl'):
            return 'ttl'
        else:
            raise ValueError(
                f'Could not determine format of ontology file {ontology_file}'
            )

    def get_nx_graph(self):
        """
        Get the networkx graph representing the ontology.
        """
        return self._nx_graph

    def get_rdf_graph(self):
        """
        Get the RDFlib graph representing the ontology.
        """
        return self._rdf_graph

    def get_root_label(self):
        """
        Get the label of the root node in the ontology.
        """
        return self._root_label

    def get_ancestors(self, node_label):
        """
        Get the ancestors of a node in the ontology.
        """
        return nx.dfs_preorder_nodes(self._nx_graph, node_label)

    def get_head_join_node(self):
        """
        Get the head join node of the ontology.
        """
        return self._head_join_node


class Ontology:
    """
    A class that represents the ontological "backbone" of a BioCypher knowledge
    graph. The ontology can be built from a single resource, or hybridised from
    a combination of resources, with one resource being the "head" ontology,
    while an arbitrary number of other resources can become "tail" ontologies at
    arbitrary fusion points inside the "head" ontology.
    """
    def __init__(
        self,
        head_ontology: dict,
        ontology_mapping: 'OntologyMapping',
        tail_ontologies: Optional[dict] = None,
    ):
        """
        Initialize the Ontology class.

        Args:
            head_ontology (OntologyAdapter): The head ontology.

            tail_ontologies (list): A list of OntologyAdapters that will be
                added to the head ontology. Defaults to None.
        """

        self._head_ontology = head_ontology
        self.extended_schema = ontology_mapping.extended_schema
        self._tail_ontology_meta = tail_ontologies

        self._tail_ontologies = None
        self._nx_graph = None

        # keep track of nodes that have been extended
        self._extended_nodes = set()

        self._main()

    def _main(self) -> None:
        """
        Main method to be run on instantiation. Loads the ontologies, joins
        them, and returns the hybrid ontology. Loads only the head ontology
        if nothing else is given. Adds user extensions and properties from
        the mapping.
        """
        self._load_ontologies()

        if self._tail_ontologies:
            for adapter in self._tail_ontologies.values():
                self._assert_join_node(adapter)
                self._join_ontologies(adapter)
        else:
            self._nx_graph = self._head_ontology.get_nx_graph()

        self._extend_ontology()

        # experimental: add connections of disjoint classes to entity
        self._connect_biolink_classes()

        self._add_properties()

    def _load_ontologies(self) -> None:
        """
        For each ontology, load the OntologyAdapter object and store it as an
        instance variable (head) or a dictionary (tail).
        """

        logger.info('Loading ontologies...')

        self._head_ontology = OntologyAdapter(
            self._head_ontology['url'],
            self._head_ontology['root_node'],
        )

        if self._tail_ontology_meta:
            self._tail_ontologies = {}
            for key, value in self._tail_ontology_meta.items():
                self._tail_ontologies[key] = OntologyAdapter(
                    value['url'],
                    value['tail_join_node'],
                    value['head_join_node'],
                )

    def _assert_join_node(self, adapter: OntologyAdapter) -> None:
        """
        Tries to find the head join node of the given ontology adapter in the
        head ontology. If the join node is not found, the method will raise an
        error.

        Args:
            adapter (OntologyAdapter): The ontology adapter of which to find the
                join node in the head ontology.
        """

        head_join_node = adapter.get_head_join_node()

        if head_join_node not in self._head_ontology.get_nx_graph().nodes:

            raise ValueError(
                f'Head join node {head_join_node} not found in '
                f'head ontology.'
            )

    def _join_ontologies(self, adapter: OntologyAdapter) -> None:
        """
        Joins the ontologies by adding the tail ontology as a subgraph to the
        head ontology at the specified join nodes.

        Args:
            adapter (OntologyAdapter): The ontology adapter of the tail ontology
                to be added to the head ontology.
        """

        if not self._nx_graph:
            self._nx_graph = self._head_ontology.get_nx_graph().copy()

        head_join_node = adapter.get_head_join_node().replace('_', ' ')
        tail_join_node = adapter.get_root_label().replace('_', ' ')
        tail_ontology = adapter.get_nx_graph()

        # subtree of tail ontology at join node
        tail_ontology_subtree = nx.dfs_tree(
            tail_ontology.reverse(), tail_join_node
        ).reverse()

        # transfer node attributes from tail ontology to subtree
        for node in tail_ontology_subtree.nodes:
            tail_ontology_subtree.nodes[node].update(tail_ontology.nodes[node])

        # rename tail join node to match head join node
        if not tail_join_node == head_join_node:
            tail_ontology_subtree = nx.relabel_nodes(
                tail_ontology_subtree, {tail_join_node: head_join_node}
            )

        # combine head ontology and tail subtree
        self._nx_graph = nx.compose(self._nx_graph, tail_ontology_subtree)

    def _extend_ontology(self) -> None:
        """
        Adds the user extensions to the ontology. Tries to find the parent in
        the ontology, adds it if necessary, and adds the child and a directed
        edge from child to parent. Can handle multiple parents.
        """

        if not self._nx_graph:
            self._nx_graph = self._head_ontology.get_nx_graph().copy()

        for key, value in self.extended_schema.items():

            if not value.get('is_a'):
                continue

            parents = _misc.to_list(value.get('is_a'))
            child = key

            while parents:
                parent = parents.pop(0)

                if parent not in self._nx_graph.nodes:

                    self._nx_graph.add_node(parent)
                    self._nx_graph.nodes[parent][
                        'label'] = _misc.sentencecase_to_pascalcase(parent)

                    # mark parent as user extension
                    self._nx_graph.nodes[parent]['user_extension'] = True
                    self._extended_nodes.add(parent)

                if child not in self._nx_graph.nodes:
                    self._nx_graph.add_node(child)
                    self._nx_graph.nodes[child][
                        'label'] = _misc.sentencecase_to_pascalcase(child)

                    # mark child as user extension
                    self._nx_graph.nodes[child]['user_extension'] = True
                    self._extended_nodes.add(child)

                self._nx_graph.add_edge(child, parent)

                child = parent

    def _connect_biolink_classes(self) -> None:
        """
        Experimental: Adds edges from disjoint classes to the entity node.
        """

        if not self._nx_graph:
            self._nx_graph = self._head_ontology.get_nx_graph().copy()

        # biolink classes that are disjoint from entity
        disjoint_classes = [
            'frequency qualifier mixin',
            'chemical entity to entity association mixin',
            'ontology class',
            'relationship quantifier',
            'physical essence or occurrent',
            'gene or gene product',
            'subject of investigation',
        ]

        for node in disjoint_classes:

            if not self._nx_graph.nodes.get(node):

                self._nx_graph.add_node(node)
                self._nx_graph.nodes[node][
                    'label'] = _misc.sentencecase_to_pascalcase(node)

            self._nx_graph.add_edge(node, 'entity')

    def _add_properties(self) -> None:
        """
        For each entity in the mapping, update the ontology with the properties
        specified in the mapping. Updates synonym information in the graph,
        setting the synonym as the primary node label.
        """

        for key, value in self.extended_schema.items():

            if key in self._nx_graph.nodes:

                self._nx_graph.nodes[key].update(value)

            if value.get('synonym_for'):

                # change node label to synonym
                if value['synonym_for'] not in self._nx_graph.nodes:
                    raise ValueError(
                        f'Node {value["synonym_for"]} not found in ontology.'
                    )

                self._nx_graph = nx.relabel_nodes(
                    self._nx_graph, {value['synonym_for']: key}
                )

    def get_ancestors(self, node_label: str) -> list:
        """
        Get the ancestors of a node in the ontology.

        Args:
            node_label (str): The label of the node in the ontology.

        Returns:
            list: A list of the ancestors of the node.
        """

        return nx.dfs_tree(self._nx_graph, node_label)

    def show_ontology_structure(self, to_disk: str = None, full: bool = False):
        """
        Show the ontology structure using treelib.
        """

        if not self._nx_graph:
            raise ValueError('Ontology not loaded.')

        if not self._tail_ontologies:
            msg = f'Showing ontology structure based on {self._head_ontology._ontology_file}'

        else:
            msg = f'Showing ontology structure based on {len(self._tail_ontology_meta)+1} ontologies: '

        print(msg)

        if not full:

            # set of leaves and their intermediate parents up to the root
            filter_nodes = set(self.extended_schema.keys())

            for node in self.extended_schema.keys():
                filter_nodes.update(self.get_ancestors(node).nodes)

            # filter graph
            G = self._nx_graph.subgraph(filter_nodes)

        else:

            G = self._nx_graph

        if not to_disk:

            # create tree
            tree = _misc.create_tree_visualisation(G)

            # add synonym information
            for node in self.extended_schema:
                if self.extended_schema[node].get('synonym_for'):
                    tree.nodes[node].tag = (
                        f'{node} = '
                        f"{self.extended_schema[node].get('synonym_for')}"
                    )

            tree.show()

            return tree

        else:

            # convert lists/dicts to strings for vis only
            for node in G.nodes:

                # rename node and use former id as label
                label = G.nodes[node].get('label')

                if not label:
                    label = node

                G = nx.relabel_nodes(G, {node: label})
                G.nodes[label]['label'] = node

                for attrib in G.nodes[label]:
                    if type(G.nodes[label][attrib]) in [list, dict]:
                        G.nodes[label][attrib] = str(G.nodes[label][attrib])

            path = os.path.join(to_disk, 'ontology_structure.graphml')

            logger.info(f'Writing ontology structure to {path}.')

            nx.write_graphml(G, path)

            return True

    def get_dict(self) -> dict:
        """
        Returns a dictionary compatible with a BioCypher node for compatibility
        with the Neo4j driver.
        """

        d = {
            'node_id': self._get_current_id(),
            'node_label': 'BioCypher',
            'properties': {
                'schema': 'self.extended_schema',
            },
        }

        return d

    def _get_current_id(self):
        """
        Instantiate a version ID for the current session. For now does simple
        versioning using datetime.

        Can later implement incremental versioning, versioning from
        config file, or manual specification via argument.
        """

        now = datetime.now()
        return now.strftime('v%Y%m%d-%H%M%S')



#!/usr/bin/env python

#
# Copyright 2021, Heidelberg University Clinic
#
# File author(s): Sebastian Lobentanzer
#                 ...
#
# Distributed under MIT licence, see the file `LICENSE`.
#
"""
BioCypher core module. Interfaces with the user and distributes tasks to
submodules.
"""
from more_itertools import peekable

from ._logger import logger

logger.debug(f'Loading module {__name__}.')

from ._write import get_writer
from ._config import config as _config
from ._config import update_from_file as _file_update
from ._create import BioCypherEdge, BioCypherNode
from ._connect import get_driver
from ._mapping import OntologyMapping
from ._ontology import Ontology
from ._translate import Translator

__all__ = ['BioCypher']

SUPPORTED_DBMS = ['neo4j']

REQUIRED_CONFIG = [
    'dbms',
    'offline',
    'strict_mode',
    'head_ontology',
]


class BioCypher:
    def __init__(
        self,
        dbms: str = None,
        offline: bool = None,
        strict_mode: bool = None,
        biocypher_config_path: str = None,
        schema_config_path: str = None,
        head_ontology: dict = None,
        tail_ontologies: dict = None,
        output_directory: str = None,
        # legacy params
        db_name: str = None,
    ):
        """
        Orchestration of BioCypher operations.

        Args:

            dbms (str): The database management system to use. For supported
                systems see SUPPORTED_DBMS.

            offline (bool): Whether to run in offline mode. If True, no
                connection to the database will be made.

            strict_mode (bool): Whether to run in strict mode. If True, the
                translator will raise an error if a node or edge does not
                provide source, version, and licence information.

            user_schema_config_path (str): Path to the user schema config
                file.

            head_ontology (dict): The head ontology defined by URL and root
                node.

            tail_ontologies (dict): The tail ontologies defined by URL and
                join nodes for both head and tail ontology.

            output_directory (str): Path to the output directory. If not
                provided, the default value 'biocypher-out' will be used.

        """

        # Update configuration if custom path is provided
        if biocypher_config_path:
            _file_update(biocypher_config_path)

        if db_name:
            logger.warning(
                'The parameter `db_name` is deprecated. Please set the '
                '`database_name` setting in the `biocypher_config.yaml` file '
                'instead.'
            )
            _config(**{'neo4j': {'database_name': db_name}})

        # Load configuration
        self.base_config = _config('biocypher')

        # Check for required configuration
        for key in REQUIRED_CONFIG:
            if key not in self.base_config:
                raise ValueError(f'Configuration key {key} is required.')

        # Set configuration - mandatory
        self._dbms = dbms or self.base_config['dbms']

        if offline is None:
            self._offline = self.base_config['offline']
        else:
            self._offline = offline

        if strict_mode is None:
            self._strict_mode = self.base_config['strict_mode']
        else:
            self._strict_mode = strict_mode

        self._schema_config_path = schema_config_path or self.base_config.get(
            'schema_config_path'
        )

        if not self._schema_config_path:
            raise ValueError(
                'BioCypher requires a schema configuration; please provide a '
                'path to the schema configuration YAML file via '
                '`biocypher_config.yaml` or `BioCypher` class parameter.'
            )

        self._head_ontology = head_ontology or self.base_config['head_ontology']

        # Set configuration - optional
        self._output_directory = output_directory or self.base_config.get(
            'output_directory'
        )
        self._tail_ontologies = tail_ontologies or self.base_config.get(
            'tail_ontologies'
        )

        if self._dbms not in SUPPORTED_DBMS:
            raise ValueError(
                f'DBMS {self._dbms} not supported. '
                f'Please select from {SUPPORTED_DBMS}.'
            )

        # Initialize
        self._ontology_mapping = None
        self._translator = None
        self._ontology = None
        self._writer = None

    def _get_ontology_mapping(self):
        """
        Create ontology mapping if not exists and return.
        """

        if not self._ontology_mapping:
            self._ontology_mapping = OntologyMapping(
                config_file=self._schema_config_path,
            )

        return self._ontology_mapping

    def _get_translator(self):
        """
        Create translator if not exists and return.
        """

        if not self._translator:
            self._translator = Translator(
                ontology_mapping=self._get_ontology_mapping(),
                strict_mode=self._strict_mode,
            )

        return self._translator

    def _get_ontology(self):
        """
        Create ontology if not exists and return.
        """

        if not self._ontology:
            self._ontology = Ontology(
                ontology_mapping=self._get_ontology_mapping(),
                head_ontology=self._head_ontology,
                tail_ontologies=self._tail_ontologies,
            )

        return self._ontology

    def _get_writer(self):
        """
        Create writer if not online.
        """

        # Get worker
        if self._offline:
            self._writer = get_writer(
                dbms=self._dbms,
                translator=self._get_translator(),
                ontology=self._get_ontology(),
                output_directory=self._output_directory,
                strict_mode=self._strict_mode,
            )
        else:
            raise NotImplementedError('Cannot get writer in online mode.')

    def _get_driver(self):
        """
        Create driver if not exists and return.
        """

        if not self._offline:
            self._driver = get_driver(
                dbms=self._dbms,
                translator=self._get_translator(),
                ontology=self._get_ontology(),
            )
        else:
            raise NotImplementedError('Cannot get driver in offline mode.')

    def write_nodes(self, nodes):
        """
        Write nodes to database.
        """

        if not self._writer:
            self._get_writer()

        nodes = peekable(nodes)
        if not isinstance(nodes.peek(), BioCypherNode):
            tnodes = self._translator.translate_nodes(nodes)
        else:
            tnodes = nodes
        # write node files
        return self._writer.write_nodes(tnodes)

    def write_edges(self, edges):
        """
        Write edges to database.
        """

        if not self._writer:
            self._get_writer()

        edges = peekable(edges)
        if not isinstance(edges.peek(), BioCypherEdge):
            tedges = self._translator.translate_edges(edges)
        else:
            tedges = edges
        # write edge files
        return self._writer.write_edges(tedges)

    def add_nodes(self, nodes):
        pass

    def add_edges(self, edges):
        pass

    def merge_nodes(self, nodes):
        """
        Merge nodes into database.
        """

        if not self._driver:
            self._get_driver()

        nodes = peekable(nodes)
        if not isinstance(nodes.peek(), BioCypherNode):
            tnodes = self._translator.translate_nodes(nodes)
        else:
            tnodes = nodes
        # write node files
        return self._driver.merge_nodes(tnodes)

    def merge_edges(self, edges):
        pass

    # OVERVIEW AND CONVENIENCE METHODS ###

    def log_missing_bl_types(self):
        """
        Get the set of Biolink types encountered without an entry in
        the `schema_config.yaml` and print them to the logger.

        Returns:
            set: a set of missing Biolink types
        """

        mt = self._translator.get_missing_biolink_types()

        if mt:
            msg = (
                'Input entities not accounted for due to them not being '
                'present in the `schema_config.yaml` configuration file '
                '(this is not necessarily a problem, if you did not intend '
                'to include them in the database; see the log for details): \n'
            )
            for k, v in mt.items():
                msg += f'    {k}: {v} \n'

            logger.info(msg)
            return mt

        else:
            logger.info('No missing Biolink types in input.')
            return None

    def log_duplicates(self):
        """
        Get the set of duplicate nodes and edges encountered and print them to
        the logger.
        """

        dn = self._writer.get_duplicate_nodes()

        if dn:

            ntypes = dn[0]
            nids = dn[1]

            msg = ('Duplicate node types encountered (IDs in log): \n')
            for typ in ntypes:
                msg += f'    {typ}\n'

            logger.info(msg)

            idmsg = ('Duplicate node IDs encountered: \n')
            for _id in nids:
                idmsg += f'    {_id}\n'

            logger.debug(idmsg)

        else:
            logger.info('No duplicate nodes in input.')

        de = self._writer.get_duplicate_edges()

        if de:

            etypes = de[0]
            eids = de[1]

            msg = ('Duplicate edge types encountered (IDs in log): \n')
            for typ in etypes:
                msg += f'    {typ}\n'

            logger.info(msg)

            idmsg = ('Duplicate edge IDs encountered: \n')
            for _id in eids:
                idmsg += f'    {_id}\n'

            logger.debug(idmsg)

        else:
            logger.info('No duplicate edges in input.')

    def show_ontology_structure(self) -> None:
        """
        Show the ontology structure of the database using the Biolink schema and
        treelib.
        """

        if not self._ontology:
            self._get_ontology()

        self._ontology.show_ontology_structure()

    def write_import_call(self) -> None:
        """
        Write a shell script to import the database depending on the chosen
        DBMS.
        """

        if not self._offline:
            raise NotImplementedError(
                'Cannot write import call in online mode.'
            )

        self._writer.write_import_call()

    # TRANSLATION METHODS ###

    def translate_term(self, term: str) -> str:
        """
        Translate a term to its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self._translator.translate_term(term)

    def reverse_translate_term(self, term: str) -> str:
        """
        Reverse translate a term from its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self._translator.reverse_translate_term(term)

    def translate_query(self, query: str) -> str:
        """
        Translate a query to its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self._translator.translate(query)

    def reverse_translate_query(self, query: str) -> str:
        """
        Reverse translate a query from its BioCypher equivalent.
        """

        # instantiate adapter if not exists
        self.start_ontology()

        return self._translator.reverse_translate(query)
